package TextData::File;

#TextData version 1.1.2 (1.2.2005)
#see POD below!

use TextData::Calc qw(LinearValue);
use TextData::Lines;

our @ISA = ('TextData::Lines');

sub new
{
	my $class = shift;
	my $config = shift;
	my $self = $class->SUPER::new($config);
	$self->{config}->{indexformat} = '%6E' unless defined $self->{config}->{indexformat};
	$self->{InFile} = shift;
	if(defined $self->{InFile})
	{
		$self->Read();
	}
	return $self;
}

sub Clear
{
	my $self = shift;
	$self->{Data} = [];
	$self->{DataIndex} = [];
	$self->{SortedData} = [];
}

sub Read
{
	my $self = shift;
	$self->Clear();
	if(defined $_[0]){ $self->{InFile} = shift; }
	if(defined $self->{InFile})
	{
		my $handle;
		if($self->{InFile} eq ''){ $handle = STDIN; }
		else{ open($handle, $self->{InFile}) or return 0; }
		my $data = 0;
		binmode($handle);
		while(<$handle>)
		{
			my $l = $_;
			if(!$data)
			{
				if($self->LineCheck(\$l)){$data = 1;}
				else{ $self->MakeNaked(\$l); push(@{$self->{RawHeader}}, \$l); }
			}
			if($data)
			{
				my $da = $self->LineData(\$l);
				push(@{$self->{Data}}, $da) if defined $da;
			}
		}
		$handle eq STDIN or close($handle);
		return 1;
	}
	else{ return 0; }
}

#ComputeIndex(\arrayofcols)
sub ComputeIndex 
{
	my $self = shift;
	my $ar = shift;
	my $zahl = $self->{config}->{numregex};
	$ar = [0] unless defined $ar;
	foreach my $c (@{$ar}){ $self->{DataIndex}->[$c] = {};}
	foreach my $d (@{$self->{Data}})
	{
		foreach my $c (@{$ar})
		{
			my $key = $d->[$c] =~ /$zahl/ ? sprintf($self->{config}->{indexformat}, $d->[$c]) : $d->[$c];
			$self->{DataIndex}->[$c]->{$key} = $d unless defined $self->{DataIndex}->[$c]->{$key};
		}
	}
}

#(col)
sub ComputeSortedIndex
{
	my $self = shift;
	my $xl = shift;
	$xl = [0] unless defined $xl;
	#compute any missing indices
	my @l = ();
	foreach my $x (@{$xl}){	push(@l, $x) unless defined $self->{DataIndex}->[$x]; }
	$self->ComputeIndex(\@l);
	#now sort all concerned indices
	foreach my $x (@{$xl})
	{
		$self->{SortedData}->[$x] = [];
		foreach my $k (sort {$a <=> $b} keys %{$self->{DataIndex}->[$x]})
		{
			push(@{$self->{SortedData}->[$x]}, $self->{DataIndex}->[$x]->{$k});
		}
	}
}


#Write(file, columns)
#maybe adding limits in future...
sub Write
{
	my $self = shift;
	$self->{OutFile} = $_[0] if defined $_[0];
	if(defined $self->{OutFile})
	{
		my $handle;
		if($self->{OutFile} eq ''){ $handle = STDOUT; }
		else{ open($handle, ">$self->{OutFile}") or return 0; }
		binmode($handle);
		my $selection = defined $_[1] ? $_[1] : undef;
		#header
		$self->WriteHeader($handle, $selection);
		#data
		for(my $i = 0; $i <= $#{$self->{Data}}; ++$i)
		{
			print $handle ${$self->DataLine($self->{Data}->[$i],$selection)};
		}
		$handle == STDOUT or close($handle);
	}
	else{ return 0; }
}

sub WriteHeader
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	foreach my $h (@{$self->{RawHeader}}){ print $handle ${$self->CommentLine($h)}; }
	if(defined $selection)
	{
		print $handle ${$self->TitleLine($selection)};			
	}
}

sub WriteNewHeader
{
	my $self = shift;
	my $handle = shift;
	my $selection = shift;
	print $handle ${$self->CommentLine(\$self->{title})};
	foreach my $c (@{$self->{comments}}){ print $handle ${$self->CommentLine(\$c)}; }
	print $handle ${$self->TitleLine($selection)};			
}

sub SetOf
{
	my $self = shift;
	my $value = shift;
	my $zahl = $self->{config}->{numregex};
	$value = sprintf($self->{config}->{indexformat}, $value) if $value =~ /$zahl/;
	my $x = shift;
	$x = 0 unless defined $x;
	defined $self->{DataIndex}->[$x] or $self->ComputeIndex([$x]);
	return $self->{DataIndex}->[$x]->{$value};	
}

sub IntSetOf
{
	my $self = shift;
	#the easy way
	my $set = $self->SetOf(@_);
	return $set if defined $set;

	#so, we got to do some Work;
	my $val = shift;
	my $x = shift;
	$x = 0 unless defined $x;
	#find suitable data value pairs for interpolation
	my $sets = $self->Neighbours($val, $x);
	return undef unless defined $sets and $#{$sets} == 1;
	#compute 
	my @set = ();
	
	for(my $y = 0; $y <= $#{$sets->[0]}; ++$y)
	{
		if($y != $x)
		{
			push(@set, LinearValue($val, [$sets->[0][$x], $sets->[1][$x]], [$sets->[0][$y],$sets->[1][$y]]));
		}
		else{ push(@set, $val); }
	}
	
	return \@set;
}

sub Y
{
	my $self = shift;
	my $val = shift;
	my $x = shift;
	my $y = shift;
	$x = 0 unless defined $x;
	$y = 1 unless defined $y;
	my $set = $self->SetOf($val,$x);
	return defined $set ? $set->[$y] : undef;
}


sub IntY
{
	my $self = shift;
	#the easy way
	my $y = $self->Y(@_);
	return $y if defined $y;

	#so, we got to do some Work;
	my $val = shift;
	my $x = shift;
	$y = shift;
	$x = 0 unless defined $x;
	$y = 1 unless defined $y;
	
	#find suitable data value pairs for interpolation
	my $sets = $self->Neighbours($val, $x);
	return undef unless defined $sets and $#{$sets} == 1;
	#compute 
	#print STDERR "Sets: @{$sets}\n";
	return LinearValue($val, [$sets->[0][$x], $sets->[1][$x]], [$sets->[0][$y],$sets->[1][$y]]);
}

sub Neighbours
{
	my $self = shift;
	my $val = shift;
	my $x = shift;
	defined $self->{SortedData}->[$x] or $self->ComputeSortedIndex([$x]);
	return undef unless $#{$self->{SortedData}->[$x]} > 0; #senseless when not at least two points there
	
	my $n = undef;
	my $e = undef;
	foreach my $v (@{$self->{SortedData}->[$x]})
	{
		$n = $e;
		$e = $v;
		if($v->[$x] > $val and defined $n){ last; }
	}
	
	return undef unless defined $n; #catches both  empty array and one-element-array
	
	return [$n,$e];
}


1;

__END__

=head1 NAME

TextData::File - process a whole file with text data

=head1 SYNOPSIS

	use TextData::File;

	#read $filename on construction
	my $file = new TextData::File(\%config,$filename);

	#create a fresh object without contents...
	my $file2 = new TextData::File(\%config);
	#...and read the file afterwards
	$file2->Read($filename);	

	print "third value of fourth data set: ",$file->{Data}->[3][2],"\n";	

=head1 DESCRIPTION

This is a subclass of TextData::Lines, so all properties are still there and only some sugar is added. It abstracts a file and contains all data for instant access (and memory consumption;-).
There is the possibility to address the data sets indexed via an arbitrary colum and to do some linear interpolation between points in this index (for that, the TextData::Calc package is needed).
A word on this index: It connects one value of the index column with the first data set it occured in - other sets containing this value are not concerned!

=head1 MEMBERS

=head2 Methods

=over * 4

=item * Clear

erases data and indexes from memory

=item * Read($file)

Read the data from the filename provided on construction or from $file if defined (where the internal value for the associated filename is set to $file). If $file is the empty string, data is read from STDIN.

=item * Write($file,\@selection)

writes the header and all data or columns in @selection to $file (use WriteFile(undef, \@selection) to write to the internally remembered output file from a previous run).

=item * WriteHeader($handle, \@selection)

writes a header to file handle $handle, tries to provide the appropriate column titles as last line when @selection is defined. Apart from this possibly constructed last line the header here is just the raw header read from the input file - including the original column titles if they were there.

=item * WriteNewHeader($handle,\@selection)

constructs a new header according to the selected columns, preserving the title and comments from the old header but putting it in current comment style and omitting the obsolete column title line, and writes it to $handle.

=item * SetOf($value,$indexcolumn) -> \@dataset

returns the data set (line) corresponding to $value in the index of $indexcolumn (0 is default if not specified)

=item * IntSetOf($value,$indexcolumn) -> \@dataset

does the same as above but applies linear interpolation to construct a data set for $value if it does not occur in the original data.

=item * Y($xvalue,$x,$y) -> $yvalue

returns a specific "Y" value to the "X" value $xvalue with the optional parameters $x and $y telling which columns we mean with X and Y - they default to 0 and 1; so the name of this method actually makes sense with files that contain X-Y data.

=item * IntY($xvalue,$x,$y) -> $yvalue

...see IntSetOf; just the same for Y()

=item * Neighbours($val,$x) -> \($n1,$n2)

searches two neighbouring points in column $x suitable for linear interpolation for $val.

=item * ComputeIndex(\@columns)

computes indexes (hash) with elements of the named columns as keys and the according data set (row) as value. Only the first found occurrence of the key in the data makes it into the index! This sets $file->{DataIndex}. 

=item * ComputeSortedIndex(\@columns)

computes the sorted indices for @columns (and the normal indices before that if necessary)

=back

=head2 Data

Here are some keys for the hash reference representing a TextData::File.

=over * 4

=item * Data

2-dimensional array reference. Stores values as $file->{Data}->[$set][$col] (data set / line and column start at 0).

=item * DataIndex

An alternative access for the data sets; $file->{DataIndex}->[$indexcol]->{$value} is a reference to the data set where $value stands in column $indexcol (in fact only the first occurence).

=item * SortedData

@{$file->{SortedData}->[$x]} contains the data sets that made it into the index of column $x in ascending order concerning the values of column $x.

=back

=head1 LICENSE

This package is open source software under the terms of the Artistic License. Free use respecting the author;-)

=head1 AUTHOR

Thomas Orgis <thomas@orgis.org>

=head1 WEB

See	http://thomas.orgis.org/scripterei for updates and lots of other (un)related stuff.
